from dataclasses import dataclass
from difflib import SequenceMatcher
import os
import yaml
import warnings

@dataclass
class YamlParser(object):

    yamlfile: str
    valid_flags: dict


    def read(self) -> None:
        self.data = self._parse_yaml()
        self._check()
        self._parse()

    def _parse_yaml(self) -> dict:
        # Retrieve raw info from yaml
        with open(self.yamlfile, 'r') as stream:
            try:
                data = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                raise(exc)
        return data


    def _check(self) -> None:
        # Check if valids in yaml file are valids
        for key in self.data.keys():
            if key not in self.valid_flags.values():
                raise KeyError(self._recommend(key))

    def _recommend(self, key):
        most_similar_flag = None
        for valid_key in self.valid_flags.values():
            flag = Most_Similar_Flag(valid_key)
            flag.calculate_distance(key)
            if not most_similar_flag:
                most_similar_flag = flag
            else:
                 if flag.distance > most_similar_flag.distance:
                     most_similar_flag = flag
        exception_raised = f"Incorrect flag {key}. Did you mean {most_similar_flag.name}?"
        return exception_raised
        
    
    def _parse(self) -> None:
        # Parse fields in yaml file and set defaults
        valid_flags = self.valid_flags
        data = self.data
        self.system = data.get(valid_flags["system"], "")
        self.system = os.path.abspath(self.system) if self.system else ""
        self.residue = data.get(valid_flags["residue"], None)
        self.chain = data.get(valid_flags["chain"], None)
        self.hbond = data.get(valid_flags["hbond"], [None, None])
        self.test = data.get(valid_flags["test"], None)
        self.pele = data.get(valid_flags["pele"], None)
        self.forcefield = data.get(valid_flags["forcefield"], "OPLS2005")
        self.verbose = data.get(valid_flags["verbose"], None)
        self.anm_freq = data.get(valid_flags["anm_freq"], None)
        self.sidechain_freq = data.get(valid_flags["sidechain_freq"], None)
        self.min_freq = data.get(valid_flags["min_freq"], None)
        self.water_freq = data.get(valid_flags["water_freq"], None)
        self.temperature = self.temp = data.get(valid_flags["temperature"], None)
        self.sidechain_resolution = data.get(valid_flags["sidechain_resolution"], None)
        self.steric_trials = data.get(valid_flags["steric_trials"], None)
        self.overlap_factor = data.get(valid_flags["overlap_factor"], None)
        self.steering = data.get(valid_flags["steering"], None)
        self.solvent = data.get(valid_flags["solvent"], None)
        self.usesrun = data.get(valid_flags["usesrun"], None)
        self.spawning = data.get(valid_flags["spawning"], None)
        self.iterations = data.get(valid_flags["iterations"], None)
        self.pele_steps = self.steps = data.get(valid_flags["pele_steps"], None)
        self.cpus = data.get(valid_flags["cpus"], None)
        self.density = data.get(valid_flags["density"], None)
        self.cluster_values = data.get(valid_flags["cluster_values"], None)
        self.cluster_conditions = data.get(valid_flags["cluster_conditions"], None)
        self.simulation_type = data.get(valid_flags["simulation_type"], None)
        self.equilibration = data.get(valid_flags["equilibration"], None)
        self.clust_type = data.get(valid_flags["clust_type"], None)
        self.eq_steps = data.get(valid_flags["eq_steps"], None)
        self.adaptive_restart = data.get(valid_flags["adaptive_restart"], None)
        self.input = data.get(valid_flags["input"], None)
        self.report_name = data.get(valid_flags["report_name"], None)
        self.traj_name = data.get(valid_flags["traj_name"], None)
        self.adaptive = data.get(valid_flags["adaptive"], None)
        self.epsilon = data.get(valid_flags["epsilon"], None)
        self.out_in = data.get(valid_flags["out_in"], None)
        self.bias_column = data.get(valid_flags["bias_column"], None)
        self.gridres = data.get(valid_flags["gridres"], 10)
        self.core = data.get(valid_flags["core"], -1)
        self.mtor = data.get(valid_flags["mtor"], 4)
        self.n = data.get(valid_flags["n"], 10000)
        self.template = data.get(valid_flags["template"], None)
        self.ext_temp = self.template
        self.rotamers = data.get(valid_flags["rotamers"], None)
        self.ext_rotamers = self.rotamers
        self.mae_lig = data.get(valid_flags["mae_lig"], None)
        self.mae_lig = os.path.abspath(self.mae_lig) if self.mae_lig else None
        self.skip_prep = self.no_ppp = data.get(valid_flags["skip_prep"], None)
        self.gaps_ter = data.get(valid_flags["gaps_ter"], None)
        self.charge_ter = data.get(valid_flags["charge_ter"], None)
        self.mpi_params = data.get(valid_flags["mpi_params"], None)
        self.nonstandard = data.get(valid_flags["nonstandard"], None)
        self.prepwizard = data.get(valid_flags["prepwizard"], None)
        self.box_center = data.get(valid_flags["box_center"], None)
        self.box_radius = data.get(valid_flags["box_radius"], None)
        self.box = data.get(valid_flags["box"], None)
        self.native = data.get(valid_flags["native"], "")
        self.atom_dist = data.get(valid_flags["atom_dist"], None)
        self.debug = data.get(valid_flags["debug"], None)
        self.folder = data.get(valid_flags["folder"], None)
        self.output = data.get(valid_flags["output"], None)
        self.randomize = data.get(valid_flags["randomize"], None)
        self.full = data.get(valid_flags["full"], None)
        self.proximityDetection = data.get(valid_flags["proximityDetection"], None)
        self.poses = data.get(valid_flags["poses"], None)
        self.precision_glide = data.get(valid_flags["precision_glide"], None) 
        self.msm = data.get(valid_flags["msm"], None)
        self.precision = data.get(valid_flags["precision"], None)
        self.clust = data.get(valid_flags["clust"], None)
        self.restart = data.get(valid_flags["restart"], None)
        self.lagtime = data.get(valid_flags["lagtime"], None)
        self.msm_clust = data.get(valid_flags["msm_clust"], None)
        self.rescoring = data.get(valid_flags["rescoring"], None)
        self.in_out = data.get(valid_flags["in_out"], None)
        self.in_out_soft = data.get(valid_flags["in_out_soft"], None)
        self.exit = data.get(valid_flags["exit"], None)
        self.exit_value = data.get(valid_flags["exit_value"], None)
        self.exit_condition = data.get(valid_flags["exit_condition"], None)
        self.exit_trajnum = data.get(valid_flags["exit_trajnum"], None)
        self.waters = data.get(valid_flags["waters"], None)
        self.water_freq = data.get(valid_flags["water_freq"], None)
        self.water_center = data.get(valid_flags["water_center"], None)
        self.water_temp = data.get(valid_flags["water_temp"], None)
        self.water_overlap = data.get(valid_flags["water_overlap"], None)
        self.water_constr = data.get(valid_flags["water_constr"], None)
        self.water_trials = data.get(valid_flags["water_trials"], None)
        self.water_radius = data.get(valid_flags["water_radius"], None)
        self.induced_fit_exhaustive = data.get(valid_flags["induced_fit_exhaustive"], None)
        self.induced_fit_fast = data.get(valid_flags["induced_fit_fast"], None)
        self.frag = data.get(valid_flags["frag"], None)
        self.ca_constr=data.get(valid_flags["ca_constr"], None)
        self.one_exit=data.get(valid_flags["one_exit"], None)
        self.box_type=data.get(valid_flags["box_type"], None)
        self.box_metric = data.get(valid_flags["box_metric"], None)
        self.time = data.get(valid_flags["time"], None)
        self.nosasa = data.get(valid_flags["nosasa"], None)
        self.sasa = data.get(valid_flags["sasa"], None)
        self.perc_sasa = data.get(valid_flags["perc_sasa"], None)
        self.seed=data.get(valid_flags["seed"], None)
        self.pdb = data.get(valid_flags["pdb"], None)
        self.log = data.get(valid_flags["log"], None)
        self.nonrenum = data.get(valid_flags["nonrenum"], None)
        self.pele_exec = data.get(valid_flags["pele_exec"], None)
        self.pele_data = data.get(valid_flags["pele_data"], None)
        self.pele_documents = data.get(valid_flags["pele_documents"], None)
        self.pca = data.get(valid_flags["pca"], None)
        self.anm_direction = data.get(valid_flags["anm_direction"], None)
        self.anm_mix_modes = data.get(valid_flags["anm_mix_modes"], None)
        self.anm_picking_mode = data.get(valid_flags["anm_picking_mode"], None)
        self.anm_displacement = data.get(valid_flags["anm_displacement"], None)
        self.anm_modes_change = data.get(valid_flags["anm_modes_change"], None)
        self.anm_num_of_modes = data.get(valid_flags["anm_num_of_modes"], None)
        self.anm_relaxation_constr = data.get(valid_flags["anm_relaxation_constr"], None)
        self.remove_constraints = data.get(valid_flags["remove_constraints"], None)
        self.pca_traj = data.get(valid_flags["pca_traj"], None)
        self.perturbation = data.get(valid_flags["perturbation"], None)
        self.binding_energy = data.get(valid_flags["binding_energy"], None)
        self.parameters = data.get(valid_flags["parameters"], None)
        self.analyse = data.get(valid_flags["analyse"], None)
        self.selection_to_perturb = data.get(valid_flags["selection_to_perturb"], None)
        self.mae = data.get(valid_flags["mae"], None)
        self.constrain_smiles = data.get(valid_flags["constrain_smiles"], None)
        self.skip_ligand_prep = data.get(valid_flags["skip_ligand_prep"], None)
        self.spawning_condition = data.get(valid_flags["spawning_condition"], None)
        self.external_constraints = data.get(valid_flags["external_constraints"], [])
        self.only_analysis = data.get(valid_flags["only_analysis"], False)
        self.overwrite = data.get(valid_flags["overwrite"], True)
        self.analysis_nclust = data.get(valid_flags["analysis_nclust"], 10)
        self.te_column = data.get(valid_flags["te_column"], 4)
        self.be_column = data.get(valid_flags["be_column"], 5)
        self.limit_column = data.get(valid_flags["limit_column"], 6)
        self.com = data.get(valid_flags["com"], None)
        self.pele_license = data.get(valid_flags["pele_license"], None)
        self.schrodinger = data.get(valid_flags["schrodinger"], None)
        self.no_check = data.get(valid_flags["no_check"], False)
        self.cleanup = data.get(valid_flags["cleanup"], False)
        self.water_empty_selector = data.get(valid_flags["water_empty_selector"], False)
        self.polarize_metals = data.get(valid_flags["polarize_metals"], False)
        self.polarization_factor = data.get(valid_flags["polarization_factor"], 2)

        # Metal constraints
        self.permissive_metal_constr = data.get(valid_flags["permissive_metal_constr"], False)
        self.constrain_all_metals = data.get(valid_flags["constrain_all_metals"], False)
        self.no_metal_constraints = data.get(valid_flags["no_metal_constraints"], False)
        
        #Frag
        self.frag_run = data.get(valid_flags["frag_run"], True)
        self.frag_core = data.get(valid_flags["frag_core"], False)
        self.frag_input = data.get(valid_flags["frag_input"], False)
        self.frag_ligands = data.get(valid_flags["frag_ligands"], False)
        self.growing_steps = data.get(valid_flags["growing_steps"], False)
        self.frag_steps = data.get(valid_flags["frag_steps"], False)
        self.frag_eq_steps = data.get(valid_flags["frag_eq_steps"], False)
        self.protocol = data.get(valid_flags["protocol"], None)
        self.frag_ai = data.get(valid_flags["frag_ai"], False)
        self.frag_ai_iterations = data.get(valid_flags["frag_ai_iterations"], False)
        self.chain_core = data.get(valid_flags["chain_core"], False)
        self.frag_restart = data.get(valid_flags["frag_restart"], False)
        self.frag_criteria = data.get(valid_flags["frag_criteria"], False)
        self.frag_output_folder = data.get(valid_flags["frag_output_folder"], False)
        self.frag_cluster_folder = data.get(valid_flags["frag_cluster_folder"], False)

        #PPI
        self.n_components = data.get(valid_flags["n_components"], None)
        self.ppi = data.get(valid_flags["ppi"], None)
        self.center_of_interface = data.get(valid_flags["center_of_interface"], None)
        self.protein = data.get(valid_flags["protein"], None)
        self.ligand_pdb = data.get(valid_flags["ligand_pdb"], None)
        self.skip_refinement = data.get(valid_flags["skip_refinement"], None)
        self.n_waters = data.get(valid_flags["n_waters"], 0)

        #Allosteric
        self.allosteric = data.get(valid_flags["allosteric"], None)
        self.skip_refinement = data.get(valid_flags["skip_refinement"], None)

        #RNA
        self.rna = data.get(valid_flags["rna"], None)

        #GPCR
        self.gpcr_orth = data.get(valid_flags["gpcr_orth"], None)
        self.orthosteric_site = data.get(valid_flags["orthosteric_site"], None)
        self.initial_site = data.get(valid_flags["initial_site"], None)

        #OUTIN
        self.final_site = data.get(valid_flags["final_site"], None)

        if self.test:
            warnings.warn("WARNING: This simulation is a test do not use the input files to run production simulations")
            self.cpus = 5
            self.pele_steps = self.steps = 1
            self.iterations = 1
            self.min_freq = 0
            self.anm_freq = 0
            self.sidechain_freq = 0
            self.n_components = 3
            self.temperature = self.temp = 10000
            self.n_components = 3

@dataclass
class Most_Similar_Flag():

    name: str

    def calculate_distance(self, key):
        self.distance = SequenceMatcher(None, self.name, key).ratio()




class BiasParams(object):


    def generate_bias_params(self, args):
        self.epsilon = args.epsilon if args.epsilon else self.simulation_params.get("epsilon", 0)
        self.bias_column = args.bias_column if args.bias_column else self.simulation_params.get("bias_column", 5)



import os
import PPP.main as ppp


# Ligand-protein input
class FragInputParameters:
    def __init__(self, args):
        self.input = args.frag_input


# If using input + series_file ligands
class FragFromCoreParameters:

    def __init__(self, args):
        self.core = args.frag_core
        self.skip_prep = args.skip_prep if args.skip_prep else self.simulation_params.get("skip_prep", False)
        if not self.skip_prep:
            self.core_process = os.path.basename(ppp.main(self.core, ".", output_pdb=["", ],
                                                          charge_terminals=args.charge_ter, no_gaps_ter=args.gaps_ter)[0])
        else:
            print("Skipping preprocessing")
            self.core_process = self.core

    @property
    def core_format(self):
        return self.core.rsplit(".")[-1]


# If using input + sdf ligands
class FragFromSDFParameters:

    def __init__(self, args):
        self.ligands = args.frag_ligands


# Base class control all input file parameters
class FragInputFiles(FragInputParameters, FragFromCoreParameters, FragFromSDFParameters):

    def __init__(self, args):
        FragInputParameters.__init__(self, args)
        FragFromCoreParameters.__init__(self, args)
        FragFromSDFParameters.__init__(self, args)



#!/usr/bin/python2.7

import os
import errno
import pandas as pd
import glob
import re
import sys
from AdaptivePELE.analysis import splitTrajectory

"""

   Description: Parse all the reports found under 'path' and sort them all
   by the chosen criteria (Binding Energy as default) having into account the
   frequency our pele control file writes a structure through the -ofreq param
   (1 by default). To sort from higher to lower value use -f "max" otherwise
   will rank the structures from lower to higher criteria's values. The number
   of structures will be ranked is controlled by -i 'nstruct' (default 10).

   For any problem do not hesitate to contact us through the email address written below.

"""

__author__ = "Daniel Soler Viladrich"
__email__ = "daniel.soler@nostrumbiodiscovery.com"

# DEFAULT VALUES
ORDER = "min"
CRITERIA = ["Binding", "Energy"]
OUTPUT = "Structure_{}.pdb"
N_STRUCTS = 10
FREQ = 1
REPORT = "report"
TRAJ = "trajectory"
ACCEPTED_STEPS = 'numberOfAcceptedPeleSteps'
OUTPUT_FOLDER = 'BestStructs'
DIR = os.path.abspath(os.getcwd())
STEPS = 3
HELP = "USE:\n\n- For xtc: python bestStructs.py 5 --top topology.pdb\n\n- For pdb:  python bestStructs.py 5"


def main(criteria, path=DIR, n_structs=10, sort_order="min", out_freq=FREQ, output=OUTPUT_FOLDER, numfolders=False, criteria2=None, topology=None, logger=None):
    """

      Description: Rank the traj found in the report files under path
      by the chosen criteria. Finally, output the best n_structs.

      Input:

         Path: Path to look for *report* files in all its subfolders.

         Criteria: Criteria to sort the structures.
         Needs to be the name of one of the Pele's report file column.
         (Default= "Binding Energy")

         n_structs: Numbers of structures to create.

         sort_order: "min" if sorting from lower to higher "max" from high to low.

         out_freq: "Output frequency of our Pele control file"

     Output:

        f_out: Name of the n output
    """
    # Get reports files
    reports = glob.glob(os.path.join(path, "*/*report*"))
    reports = glob.glob(os.path.join(path, "*report*")) if not reports else reports
    reports = filter_non_numerical_folders(reports, numfolders)
    try:
        reports[0]
    except IndexError:
        raise IndexError("Not report file found. Check you are in adaptive's or Pele root folder")

    # Get metrics of reports
    if criteria.isdigit():
        steps, criteria = get_column_names(reports, STEPS, criteria)
    else:
        steps = get_column_names(reports, STEPS, criteria)

    # Data Mining
    if not criteria2:
        min_values = parse_values(reports, n_structs, criteria, sort_order, steps)
    else:
        min_values = parse_values(reports, n_structs, criteria, criteria2, sort_order, steps)

    # Process data to be outputted
    values = min_values[criteria].tolist()
    paths = min_values[DIR].tolist()
    epochs = [os.path.basename(os.path.normpath(os.path.dirname(Path))) for Path in paths]
    file_ids = min_values.report.tolist()
    step_indexes = min_values[steps].tolist()
    files_out = ["epoch{}_trajectory_{}.{}_{}{}.pdb".format(epoch, report, int(step), criteria.replace(" ",""), value) \
       for epoch, step, report, value in zip(epochs, step_indexes, file_ids, values)]

    # Read trajectory and output snapshot
    for f_id, f_out, step, path in zip(file_ids, files_out, step_indexes, paths):
        if not topology:
            extract_snapshot_from_pdb(path, f_id, output, topology, step, out_freq, f_out, logger=logger)
        else:
            extract_snapshot_from_xtc(path, f_id, output, topology, step, out_freq, f_out, logger=logger)
    files_out = [os.path.join(output, f) for f in files_out]
    # Return data
    return files_out, epochs, file_ids, step_indexes, values


def extract_snapshot_from_pdb(path, f_id, output, topology, step, out_freq, f_out, logger=None):
    f_in = glob.glob(os.path.join(os.path.dirname(path), "*trajectory*_{}.pdb".format(f_id)))
    if not f_in:
        f_in = glob.glob(os.path.join(os.path.dirname(path), "*trajectory*_{}.*".format(f_id)))
    if len(f_in) == 0:
        sys.exit("Trajectory {} not found. Be aware that PELE trajectories must contain the label \'trajectory\' in their file name to be detected".format("*trajectory*_{}".format(f_id)))
    f_in = f_in[0]
    with open(f_in, 'r') as input_file:
        file_content = input_file.read()
    trajectory_selected = re.search('MODEL\s+%d(.*?)ENDMDL' % int((step) / out_freq+1), file_content,re.DOTALL)

    # Output Snapshot
    try:
        mkdir_p(output)
    except OSError:
        pass

    traj = []
    with open(os.path.join(output, f_out), 'w') as f:
        traj.append("MODEL     %d" % int((step)/out_freq+1))
        try:
            traj.append(trajectory_selected.group(1))
        except AttributeError:
            raise AttributeError("Model not found. Check the -f option.")
        traj.append("ENDMDL\n")
        f.write("\n".join(traj))


def extract_snapshot_from_xtc(path, f_id, output, topology, step, out_freq, f_out, logger=None):
    f_in = glob.glob(os.path.join(os.path.dirname(path), "*trajectory*_{}.xtc".format(f_id)))
    if not f_in:
        f_in = glob.glob(os.path.join(os.path.dirname(path), "*trajectory*_{}.*".format(f_id)))
    if len(f_in) == 0: 
        sys.exit("Trajectory {} not found. Be aware that PELE trajectories must contain the label \'trajectory\' in their file name to be detected".format("*trajectory*_{}".format(f_id)))
    splitTrajectory.main(output, [f_in[0], ], topology, [(step)/out_freq+1, ], template= f_out)
    logger.info("Model {} selected".format(f_out))


def parse_values(reports, n_structs, criteria, sort_order, steps):
    """

       Description: Parse the 'reports' and create a sorted array
       of size n_structs following the criteria chosen by the user.

    """

    INITIAL_DATA = [(DIR, []),
                    (REPORT, []),
                    (steps, []),
                    (criteria, [])
                    ]
    min_values = pd.DataFrame.from_dict(INITIAL_DATA)
    for file in reports:
        report_number = os.path.basename(file).split("_")[-1]
        data = pd.read_csv(file, sep='    ', engine='python')
        selected_data = data.loc[:, [steps, criteria]]
        if sort_order == "min":
                report_values = selected_data.nsmallest(n_structs, criteria)
                report_values.insert(0, DIR, [file]*report_values[criteria].size)
                report_values.insert(1, REPORT, [report_number]*report_values[criteria].size)
                mixed_values = pd.concat([min_values, report_values])
                min_values = mixed_values.nsmallest(n_structs, criteria)

        else:
                report_values = selected_data.nlargest(n_structs, criteria)
                report_values.insert(0, DIR, [file]*report_values[criteria].size)
                report_values.insert(1, REPORT, [report_number]*report_values[criteria].size)
                mixed_values = pd.concat([min_values, report_values])
                min_values = mixed_values.nlargest(n_structs, criteria)
    return min_values


def filter_non_numerical_folders(reports, numfolders):
    """
    Filter non numerical folders among
    the folders to parse
    """
    if numfolders:
        new_reports = [report for report in reports if(os.path.basename(os.path.dirname(report)).isdigit())]
        return new_reports
    else:
        return reports


def get_column_names(reports, steps, criteria):
    data = pd.read_csv(reports[0], sep='    ', engine='python')
    data = list(data)
    if criteria.isdigit():
        return data[int(steps)-1], data[int(criteria)-1]
    else:
        return data[int(steps)-1]


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
import versioneer

here = path.abspath(path.dirname(__file__))
ext_modules = []
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
     long_description = f.read()

setup(
    name="pele_platform",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Automatic platform to launch PELE',
    long_description=long_description,
    url="https://github.com/NostrumBioDiscovery/pele_platform",
    author='Daniel Soler',
    author_email='daniel.soler@nostrumbiodiscovery.com',
    packages=find_packages(exclude=['docs', 'tests', 'tests.data']),
    package_data={"pele_platform/AdaptivePELE/atomset": ['*.pxd'], "pele_platform/AdaptivePELE/freeEnergies/": ['*.pyx']},
    include_package_data=True,
    install_requires=["AdaptivePELE", "frag_pele==2.2.1", "lib_prep", "PPPele==1.0.5", "PlopRotTemp==1.0.1", "PyYAML", "fpdf", "scikit-learn>0.2", "pillow", "scipy", "matplotlib", 
       "biopython", "pandas", "pytest", "cython", "numpy"],    
    ext_modules=ext_modules  # accepts a glob pattern
)



from dataclasses import dataclass
import os
import AdaptivePELE.adaptiveSampling as ad
from pele_platform.Utilities.Helpers import helpers, template_builder
import pele_platform.Utilities.Parameters.pele_env as pv
import pele_platform.Utilities.Helpers.water as wt

@dataclass
class SimulationBuilder(template_builder.TemplateBuilder):

    adaptive_file: str
    pele_file: str
    topology: str

    def generate_inputs(self, env: pv.EnviroBuilder, water_obj: wt.WaterIncluder) -> None:
        # Fill in simulation template with 
        # simulation parameters specified in  
        # pv.EnviroBuilder
        env.logger.info("Running Simulation")
        # Fill to time because we have flags inside flags
        self.fill_pele_template(env, water_obj)
        self.fill_pele_template(env, water_obj)
        self.fill_adaptive_template(env)
        self.fill_adaptive_template(env)
        if not env.debug:
            self.run()
        env.logger.info("Simulation run successfully (:\n\n")

    def fill_pele_template(self, env: pv.EnviroBuilder, water_obj: wt.WaterIncluder) -> None:
        # Fill in PELE template
        self.pele_keywords = { "PERTURBATION": env.perturbation, "SELECTION_TO_PERTURB": env.selection_to_perturb,
                        "BE": env.binding_energy, "SASA": env.sasa,
                        "LOGFILE": env.logfile, "NATIVE": env.native, "FORCEFIELD": env.forcefield, "CHAIN": env.chain, 
                        "CONSTRAINTS": "\n".join(env.constraints), "CPUS":env.cpus,
                        "LICENSES": env.license, "BOX_RADIUS": env.box_radius, "BOX_CENTER": env.box_center,
                        "SASA_min": env.sasa_min, "SASA_max": env.sasa_max,
                        "WATER_RADIUS": water_obj.water_radius, "WATER_CENTER": water_obj.water_center, "WATER": water_obj.water_line,
                        "WATER_ENERGY": water_obj.water_energy, "METRICS": env.metrics, "REPORT_NAME": env.report_name, 
                        "TRAJ_NAME": env.traj_name,
                        "SOLVENT": env.solvent, "PARAMETERS": env.parameters, "SIDECHAIN_RESOLUTION": env.sidechain_resolution,
                        "OVERLAP": env.overlap_factor, "STERIC_TRIALS": env.steric_trials, "TEMPERATURE": env.temperature, 
                        "MIN_FREQ": env.min_freq, "SIDECHAIN_FREQ": env.sidechain_freq, "ANM_FREQ": env.anm_freq, "BOX" : env.box, "PROXIMITY": env.proximityDetection, "WATER_FREQ": env.water_freq, "VERBOSE": env.verbose, "ANM_DISPLACEMENT": env.anm_displacement, "ANM_MODES_CHANGE": env.anm_modes_change, "ANM_DIRECTION": env.anm_direction, "ANM_MIX_MODES": env.anm_mix_modes, "ANM_PICKING_MODE": env.anm_picking_mode,
                        "ANM_NUM_OF_MODES": env.anm_num_of_modes, "ANM_RELAXATION_CONST": env.anm_relaxation_constr,
                        "PCA": env.pca, "COMPLEXES": env.complexes, "PELE_STEPS": env.frag_pele_steps, "OUTPUT_PATH": env.output_path,
                        "COM": env.com, "STEERING": env.steering}


        super(SimulationBuilder, self).__init__(self.pele_file, self.pele_keywords)

    def fill_adaptive_template(self, env: pv.EnviroBuilder) -> None:
        # Fill in adaptive template
        self.adaptive_keywords = { "RESTART": env.adaptive_restart, "OUTPUT": env.output, "INPUT":env.adap_ex_input,
                "CPUS":env.cpus, "PELE_CFILE": os.path.basename(self.pele_file), "LIG_RES": env.residue, "SEED": env.seed, "EQ_STEPS": env.equil_steps,
                "EQUILIBRATION":env.equilibration, "EPSILON": env.epsilon, "BIAS_COLUMN": env.bias_column, "ITERATIONS": env.iterations, 
                "PELE_STEPS": env.pele_steps, "REPORT_NAME": env.report_name, "SPAWNING_TYPE": env.spawning, "DENSITY": env.density,
                "SIMULATION_TYPE": env.simulation_type, "CLUSTER_VALUES": env.cluster_values, "CLUSTER_CONDITION": env.cluster_conditions,
                "UNBINDING": env.unbinding_block, "USESRUN": env.usesrun, "LIGAND": env.ligand,
                "PELE_BIN": env.pele_exec, "PELE_DATA": env.pele_data, "PELE_DOCUMENTS": env.pele_documents,
                "CONDITION": env.spawning_condition, "MPI_PARAMS": env.mpi_params, "CLUST_TYPE": env.clust_type}
        super(SimulationBuilder, self).__init__(self.adaptive_file, self.adaptive_keywords)


    def run(self, hook=False) -> None:
        # Launch montecarlo simulation
        with helpers.cd(os.path.dirname(self.adaptive_file)):
                ad.main(self.adaptive_file)




#!/usr/bin/env python
'''
Created in July, 2014
Author: Christoph Grebner
E-Mail: christoph.grebner@astrazeneca.com

program to calculate ANM and PCA for a given set of proteins in PDB format
uses the ProDy library 

'''
import prody
import glob
import argparse
import numpy as np


class PCA_ANM_Analysis:
    '''
    class definition for PCA and ANM analysis using ProDy library
    calculate the PCA of a given set of structures and the ANM of the reference structure
    '''
    def __init__(self, selection="calpha"):
        #Initialization of variables with default values
        self.selection=selection
        self.vmd=False
        self.compare=False
        self.ref="none"
        self.debug=False
        self.pdbarg="all"

        self.pdbs = []
 
    def do_Auto_analysis(self):
        pdbs = self.getPDBs()
        ensemble = self.createEnsemble(pdbs)
        pca = self.calcPCA(ensemble)
        anm = self.calcANM(ensemble.getConformation(0))
        if self.compare:
            self.compare_modes(anm,pca)
    
    def printSettings(self):
        '''
        Print all settings for debugging purposes
        '''
        env.logger.info("Settings in PCA_ANM_Analysis")
        env.logger.info("selection, VMD, compare, ref, pdb")
        env.logger.info(self.selection, self.vmd, self.compare, self.ref, self.pdbarg)


    def set_default_based_on_argparse(self, selection, vmd, compare, ref, debug, pdb):
        '''
        set arguments from command line based on argparser
        '''
        if selection in ["calpha", "backbone", "all"]:
            self.selection=selection
        else:
            env.logger.info("Invalid value for selection:", selection)
            env.logger.info("Possible Values for selection are: calpha, backbone, all")
            exit()
        self.vmd = vmd
        self.compare=compare
        self.ref=ref
        self.debug=debug
        self.pdbarg=pdb
        
        #set prody level of output
        if self.debug:
            prody.confProDy(verbosity="debug")
        else:
            prody.confProDy(verbosity="info")
            
        if self.debug:
            env.logger.info("***DEBUG***")
            self.printSettings()
            env.logger.info("***DEBUG-END***")
    
    def getPDBs(self):
        '''
        Evaluate information f#from PCA_analysis_GUI import *
from PCA_dPCA import *
rom command line
        Get PDBs and return array
        '''
        if self.pdbarg == "all":
            pdbs = glob.glob('*.pdb')
        else:
            pdbs = self.pdbarg.split()
            
        if self.debug:
            env.logger.info("***DEBUG*** The chosen PDBs are:", pdbs)
        self.pdbs = pdbs    
        return pdbs

        
    def createEnsemble(self, pdbs, logger):
        '''
        Create a prody ensemble based on getPDBs return
        Take into account, that system can be prepared or not
        and take gaps and multiple chains into account
        take single files and trajectory file into account
        
        pdbs: list-array with pdb-filenames
        
        ToDo: 
            - set reference depending on longest chain with no gaps
            - check for duplicate chains and only select this one
        '''
        logger.info("Create Ensemble")
        ref_chids = []
        
        #set the reference structure, if not chosen by user, the first Frame or PDB is taken
        #remove reference Frame/PDB from list
        if self.ref == "none":
            self.ref = pdbs[0]
            pdbs.pop(0)
        else:
            if self.ref in pdbs:
                if self.debug:
                    logger.info("***DEBUG*** Found Ref:", self.ref)
                    logger.info("***DEBUG*** Index Ref", pdbs.index(self.ref))
                    logger.info("Removing reference from pdblist")
                pdbs.pop(pdbs.index(self.ref))
   
        if self.debug:
            logger.info("***DEBUG*** Reference is",self.ref)
            
        #open reference file and check for HID, HIE; if found, replace by HIS
        try:
            pdbinfile = open(self.ref)   
        except IOError as e:
            env.logger.info(self.ref)
            env.logger.info("I/O error({0}): {1}".format(e.errno, e.strerror))
            exit()
        
        pdbfiledata = pdbinfile.read()
        for line in pdbinfile:
            if line[17:20] == "HID" or line[17:20] == "HIE":
                env.logger.info("HID or HIE found in pdb-structure. The names will be changed to HIS in the input file")
                break
        pdbinfile.close()
        
        pdbfiledata = pdbfiledata.replace("HID","HIS")
        pdbfiledata = pdbfiledata.replace("HIE","HIS")
        
        logger.info("Warning!!! Rewriting reference file {}".format(self.ref))
        pdbfile = open(self.ref,'w')
        pdbfile.write(pdbfiledata)
        pdbfile.close()
        
        #set reference and create first structure  
        if self.selection == "all":
            ref_structure = prody.parsePDB(self.ref)
        else:
            ref_structure = prody.parsePDB(self.ref, subset=self.selection)
    

        reference_hierview = ref_structure.getHierView()
        
        logger.info(f"Found {reference_hierview.numChains()} Chain(s) in {reference_hierview._atoms.getTitle()}")
        ensemble_ref_title = reference_hierview._atoms.getTitle()
        logger.info(reference_hierview[0])
        
        '''
        at the moment all chains are taken
        --> check for duplicate chains and only take on of those
        '''
        
        for chain in reference_hierview:
            ref_chids.append(chain.getChid())
        if self.debug:
            env.logger.info(f"***DEBUG*** {ref_chids}")
        
        reference_chains = [reference_hierview[chid] for chid in ref_chids] 
        
        if self.debug:
            env.logger.info(f"***DEBUG*** {reference_chains}")
            
        
        
        ref_chain = reference_chains[0]
        for i in range (1, len(reference_chains), 1):
            ref_chain = ref_chain + reference_chains[i]
        if self.debug:
            env.logger.info("***DEBUG***", ref_chain)
        # save globally
        self.ref_chain = ref_chain
        #Create Ensemble of structures
        ensemble = prody.PDBEnsemble(ensemble_ref_title)
        ensemble.setAtoms(ref_chain)
        ensemble.setCoords(ref_chain)
        #Set ref_structure as first coordinate set
        ensemble.addCoordset(ref_structure)
        
        if self.debug:
            env.logger.info("***DEBUG***", ref_chain.getResnames())
            env.logger.info("***DEBUG***", ref_chain.getResnums())
        
        unmapped = []
        # map remaining structures to reference chain and add to ensemble if mapped
        for pdb in pdbs:
             if self.debug:
                 env.logger.info("***DEBUG*** Processing ", pdb)
             if self.selection == "all":
                 #structure = prody.parsePDB(self.ref)
                 structure = prody.parsePDB(pdb)
             else:
                 structure = prody.parsePDB(pdb, subset=self.selection)
             
             atommaps = []
             for reference_chain in reference_chains:
                 # Map current PDB file to the reference chain
                 mappings = prody.mapOntoChain(structure, reference_chain,
                                         seqid=90,
                                         coverage=50,
                                         subset=self.selection)

                 if len(mappings) == 0:
                     env.logger.info('Failed to map', pdb)
                     break
                 atommaps.append(mappings[0][0])
                 # Make sure all chains are mapped
             if len(atommaps) != len(reference_chains):
                 unmapped.append(pdb)
                 continue
             #env.logger.info(atommaps[0]
             atommap = atommaps[0]
             for i in range (1, len(reference_chains), 1):
                  atommap = atommap + atommaps[i]
             # Add the atommap (mapped coordinates) to the ensemble
             # Note that some structures do not completely map (missing residues)
             # so we pass weights (1 for mapped atoms, 0 for unmapped atoms)
             ensemble.addCoordset(atommap, weights=atommap.getFlags('mapped'))
     
        if self.debug:
            env.logger.info("***DEBUG*** Number of conformations:", ensemble.numConfs())
            env.logger.info("***DEBUG*** ", ensemble)
            env.logger.info("***DEBUG*** Unmapped structures:", unmapped)
        ensemble.iterpose()

        #env.logger.info(ensemble.getTitle()
        return ensemble
    
    def setSelections(self, ensemble):
        '''
        put selections also into key arguments
        '''        
        
        custom_selection = False
        
        if custom_selection == False:
            self.selection_ref_structure = self.ref_chain
            return ensemble
        else:
            self.selection2="name CA"
            structure =  ensemble.getAtoms().select(self.selection2)
            #selec = structure.getSelstr().split()
            selec = structure.getIndices()
            if self.debug:
                env.logger.info("***DEBUG***", selec)
                env.logger.info("***DEBUG***", len(selec))
        
            #create new ensemble containing only the selected atoms        
            ensemble2 = prody.PDBEnsemble(ensemble.getTitle())
            ensemble2.setAtoms(structure)
            ensemble2.setCoords(structure.getCoords())
            self.selection_ref_structure = structure
            
            #ensemble2.addCoordset(structure)                
            for coords in ensemble.iterCoordsets():
                new_coords = np.zeros((ensemble2.numSelected(), 3))
                count = 0
                for coord in range(len(coords)):
                    if coord in selec:
                        new_coords[count] = coords[coord]
                        #env.logger.info(count, new_coords[count]
                        count+=1
                ensemble2.addCoordset(new_coords) 
                
            ensemble2.getConformation(0).setLabel(ensemble2.getTitle())    

            if self.debug:
                env.logger.info("***DEBUG*** Ensemble before selections: ", repr(ensemble))
                env.logger.info("***DEBUG*** Ensemble after selections: ", repr(ensemble2))
            
            return ensemble2
    
    def setWeights(self, ensemble):
        '''
        create a custom selection (e.g. only specific residues) and set the weights for the remaining atoms
        to 0
        --> They will not contribute to the PCA
        the original ensemble is not changed!
        '''
        env.logger.info("Set weights")
        self.selection2="(backbone)"
        structure =  ensemble.getAtoms().select(self.selection2)
        #selec = structure.getSelstr().split()
        selec = structure.getIndices()
        #set weights for atoms outside selection to zero!
        weights =  ensemble.getWeights()
        for i in range(1,len(weights)):
            for j in range(len(weights[i])):               
                #set atoms in selection to 1 rest to 0
                if j not in selec:
                    weights[i][j] = 0
                #env.logger.info(j, weights[i][j]
        
        ensemble.setWeights(weights)

        return ensemble
        
    def calcPCA(self, ensemble, logger):
        '''
        calcPCA:
        #ensemble: prody ensmeble with structure information
        
        calculate PCA for a set of structures    
        
        return: prody.pca object
        '''
        logger.info("Calculate PCA")
             
        PCAname = ensemble.getTitle()
        pca = prody.PCA(PCAname)
        pca.buildCovariance(ensemble)

        logger.info("PCA")
        pca.calcModes()
        logger.info(repr(pca))
        
        outputname = PCAname + "_pca_modes.nmd"
                    
        prody.writeNMD(outputname, pca[:10], self.selection_ref_structure)
        
        if self.vmd == True:
            prody.viewNMDinVMD(outputname)
        logger.info(f"PCA is saved in: {outputname}")
        return pca, outputname
        
    def calcANM(self, structure, logger):
        '''
        calcANM:
        #structure: prody PDB-structure
        
        calculate ANM for one specific structure    
        
        return: prody.anm object
        '''
        logger.info("Calculate ANM")
        ANMname = structure.getLabel()
        anm = prody.ANM(ANMname)
        anm.buildHessian(structure, cutoff=15.0)
        anm.calcModes()
        # write out the three lowest modes as NMD file (visualize with NMWizard in VMD
        outputname = ANMname + "_anm_modes.nmd"
        prody.writeNMD(outputname, anm[:10], self.selection_ref_structure)
        if self.vmd == True:
            prody.viewNMDinVMD(outputname)
        logger.info(f"ANM is saved in: {outputname}")
        return anm

    def compare_modes(self, modes1, modes2):
        '''
        Compare the calculated ANM and PCA
        '''
        
        env.logger.info(self.print_Eigen_Val_Vec(modes1))
        env.logger.info(self.print_Eigen_Val_Vec(modes2))
        
        prody.printOverlapTable(modes2[:3], modes1[:3]) # Top 3 PCs vs slowest 3 ANM modes
        
        for mode in modes2[:10]:
            var = prody.calcFractVariance(mode).round(3)
            env.logger.info('{0:s}  % variance = {1:.2f}'.format(mode, var))
            coll = prody.calcCollectivity(mode)
            env.logger.info('{0:s}  collectivity = {1:.2f}'.format(mode, coll))
    
        for mode in modes1[:10]:    # Print ANM mode collectivity
            var = prody.calcFractVariance(mode).round(3)
            env.logger.info('{0:s}  % variance = {1:.2f}'.format(mode, var))
            coll = prody.calcCollectivity(mode)
            env.logger.info('{0:s}  collectivity = {1:.2f}'.format(mode, coll))
        
    def print_Eigen_Val_Vec(self,mode):
        env.logger.info("Eigenvalues of", mode.getModel())
        env.logger.info(mode.getEigvals().round(3))
        env.logger.info("Eigenvectors of", mode.getModel())
        env.logger.info(mode.getEigvecs().round(3))

    def print_Match(self, match):
       env.logger.info('Chain 1     : {}'.format(match[0]))
       env.logger.info('Chain 2     : {}'.format(match[1]))
       env.logger.info('Length      : {}'.format(len(match[0])))
       env.logger.info('Seq identity: {}'.format(match[2]))
       env.logger.info('Seq overlap : {}'.format(match[3]))
       env.logger.info('RMSD        : {}\n'.format(calcRMSD(match[0], match[1])))

    def print_modes(self, modes):
        env.logger.info("Printing modes to file")
        import matplotlib.pyplot as plt
        residues = self.ref_chain.getResnums()
        print_offset = residues[0]
        plt.rcParams.update({'font.size': 14})

        for count in range(len(modes)):
            #array with lenghts of vectors
            a3d = (modes[count].getArrayNx3()**2).sum(axis=1)**0.5
            plt.plot(a3d[:])
            plt.xlabel('Residue index')
            plt.ylabel('Lenght of fluctuation vector')         
            locs,labels = plt.xticks()
            new_labels = ['%d' % (a+print_offset) for a in locs]
            plt.xticks(locs,new_labels)
            plt.title(str(modes)+" mode "+str(count+1))
            plt.savefig(str(modes)+"_mode_"+str(count)+".pdf")
            plt.clf()   
        
        for count in range(6):
            #array with lenghtes of vectors
            a3d = (modes[count].getArrayNx3()**2).sum(axis=1)**0.5
            plt.plot(a3d[:], label=("Mode "+str(count+1)))

        plt.xlabel('Residue index')
        plt.ylabel('Lenght of fluctuation vector')
        
        locs,labels = plt.xticks()
        new_labels = ['%d' % (a+print_offset) for a in locs]
        plt.xticks(locs,new_labels)
        lgd = plt.legend(loc='center right', bbox_to_anchor=(1.3,0.5))
        plt.savefig(str(modes)+"_mode_1_to_6.pdf", bbox_extra_artists=(lgd, ), bbox_inches='tight')
        
        env.logger.info("Printing modes to file finished")

          
    def print_diff(self, pca_modes, anm_modes):
        env.logger.info("here I am in DIFF")
        import matplotlib.pyplot as plt
        from matplotlib.path import Path
        import matplotlib.patches as patches
        residues = self.ref_chain.getResnums()
        print_offset = residues[0]
        #Paper01:PCA-X-Ray:
        print_offset = 738
        plt.rcParams.update({'font.size': 24}) 
        plt.rcParams['xtick.major.pad'] = 10
        plt.rcParams['ytick.major.pad'] = 10
        
        a3d_pca = ((pca_modes[0].getArrayNx3())**2).sum(axis=1)**0.5
        a3d_anm = ((anm_modes[0].getArrayNx3())**2).sum(axis=1)**0.5
        a3d_all = a3d_pca - a3d_anm
        for count in range(1,6):
            #array with lenghts of vectors
            a3d_pca = ((pca_modes[count].getArrayNx3())**2).sum(axis=1)**0.5
            a3d_anm = ((anm_modes[count].getArrayNx3())**2).sum(axis=1)**0.5
            a3d_all = a3d_all + a3d_pca - a3d_anm
            
            
        plt.plot(a3d_all[:], label=("Difference"), lw=1.5)
        plt.xlabel('Residue index')
        plt.ylabel('|PCA| - |ANM|')         
        locs,labels = plt.xticks()
        new_labels = ['%d' % (a+print_offset) for a in locs]
        plt.xticks(locs,new_labels)
        lgd = plt.legend(loc='center right', bbox_to_anchor=(1.65,0.5))
        ##add boxes
        codes = [Path.MOVETO,
                  Path.LINETO,
                  Path.LINETO,
                  Path.LINETO,
                  Path.CLOSEPOLY,
                  ]
        ### light green
        verts = [
                 (95, -2.0), # left, bottom
                 (95, 2.0), # left, top
                 (115, 2.0), # right, top
                 (115, -2.0), # right, bottom
                 (0., 0.), # ignored
                 ]
        path = Path(verts, codes)
        #add box around area of interest
        ax = plt.subplot()
        patch = patches.PathPatch(path, facecolor='#CCFFCC', lw=1)
        #patch = patches.PathPatch(path, facecolor='lightgreen', lw=1)
        ax.add_patch(patch)
        plt.axhline(y=0.0,xmin=0,xmax=3,c="black",linewidth=1.0,zorder=0)
        plt.savefig(str(pca_modes)+"_all_DIFF.pdf", bbox_extra_artists=(lgd, ), bbox_inches='tight')
        plt.clf()   
        
        
        for count in range(6):
            #array with lenghtes of vectors         
            a3d = ((pca_modes[count].getArrayNx3())**2).sum(axis=1)**0.5 - ((anm_modes[count].getArrayNx3())**2).sum(axis=1)**0.5         
            plt.plot(a3d[:], label=("Mode "+str(count+1)), lw=1.5)
        plt.xlabel('Residue index')
        plt.ylabel('|PCA| - |ANM|')   
        locs,labels = plt.xticks()
        new_labels = ['%d' % (a+print_offset) for a in locs]
        plt.xticks(locs,new_labels)
        lgd = plt.legend(loc='center right', bbox_to_anchor=(1.65,0.5))
        
        ##add boxes
        codes = [Path.MOVETO,
                  Path.LINETO,
                  Path.LINETO,
                  Path.LINETO,
                  Path.CLOSEPOLY,
                  ]
        ### light green
        verts = [
                 (95, -0.7), # left, bottom
                 (95, 0.7), # left, top
                 (115, 0.7), # right, top
                 (115, -0.7), # right, bottom
                 (0., 0.), # ignored
                 ]
        
        path = Path(verts, codes)
        #add box around area of interest
        ax = plt.subplot()
        patch = patches.PathPatch(path, facecolor='#CCFFCC', lw=1)
        ax.add_patch(patch)
        plt.axhline(y=0.0,xmin=0,xmax=3,c="black",linewidth=1.0,zorder=0)
        plt.savefig(str(pca_modes)+"_ANM_PCA_DIFF.pdf", bbox_extra_artists=(lgd, ), bbox_inches='tight')
        
    def print_from_file(self, modesfile):
        env.logger.info("Printing modes from file", modesfile)
        import matplotlib.pyplot as plt
        from matplotlib.path import Path
        import matplotlib.patches as patches
        modes, atm_group = prody.parseNMD(modesfile)

        title=modesfile.split(".")
        modes.setTitle(title[0])
        print_offset = 1
        #print_offset = 738
        plt.rcParams.update({'font.size': 24}) 
        plt.rcParams['xtick.major.pad'] = 10
        plt.rcParams['ytick.major.pad'] = 10
        
        for count in range(6):
            #array with lenghtes of vectors
            a3d = (modes[count].getArrayNx3()**2).sum(axis=1)**0.5
            plt.plot(a3d[:], label=("Mode "+str(count+1)))
        plt.xlabel('Residue index')
        plt.ylabel('Lenght of fluctuation vector')
        
        locs,labels = plt.xticks()
        new_labels = ['%d' % (a+print_offset) for a in locs]
        plt.xticks(locs,new_labels)
        plt.ylim([0,0.8])
        plt.xlim([print_offset, modes.numAtoms()])
        
        lgd = plt.legend(loc='center right', bbox_to_anchor=(1.6,0.5))
        plt.savefig(str(modes.getTitle())+"_mode_1_to_6.pdf", bbox_extra_artists=(lgd, ), bbox_inches='tight')
        
        env.logger.info("Printing modes to file finished")

def parse_Arguments():
    '''
    parsing of command line arguments
    '''
    parser = argparse.ArgumentParser(
        description='Calculate PCA and ANM for given structures',
        epilog="", formatter_class=SmartFormatter)
    parser.add_argument("--pdb", help="""R|specify the pdbs which should be used. 
            Important!!!: They will be overwritten by Prody to match its PDB format,
            in particular histidine names.
            If not specified, all PDBs in the current folder are taken by default.
            Possibilities:
            use all PDBs in directory: --pdb all
            give a list of PDBs, either single or trajectory file: --pdb "name1.pdb name2.pdb name3.pdb name4.pdb"
            """, default='all')
    parser.add_argument("--selection", help="""R|Specify the selection for normal mode analysis
            valid options:
            --selection "calpha": 
            --selection "backbone"
            --selection "all"
    """, default='calpha')
    parser.add_argument("--compare", help="compare PCA and ANM", action="store_true", default=False)
    parser.add_argument("--vmd", help="Visualize result in VMD", action="store_true", default=False)
    parser.add_argument("--debug", help="Print debug information", action="store_true", default=False)
    parser.add_argument("--ref", help="Select a reference structure for PCA and ANM", default='none')
    parser.add_argument("--print_mode", help="Print the square fluctuations of the modes", action="store_true", default=False)
    parser.add_argument("--print_diff", help="Print difference between ANM and PCA modes", action="store_true", default=False)
    parser.add_argument("--print_file", help="Print modes from a NMD file. This will exit the program after printing", default='none')
    parser.add_argument("--anm_only", help="Only perform ANM analysis", action="store_true", default=False)
    return parser.parse_args()


class SmartFormatter(argparse.HelpFormatter):
    """
    Helper class for argparse
    help beginning with R| will be parsed with new lines
    """
    def _split_lines(self, text, width):
        # this is the RawTextHelpFormatter._split_lines
        if text.startswith('R|'):
            return text[2:].splitlines()
        return argparse.HelpFormatter._split_lines(self, text, width)


def main(pdb, selection='calpha', vmd=False, compare=False, ref="none", debug=False, logger=None):

    logger.info("\n\nWarning!!! Reference PDB file, if provided, will be rewritten in"
          " ProDy PDB format, with HID and HIE residues renamed to HIS.\n\n")
    pdb = " ".join(pdb)
        
    # create PCA_analysis object
    Analysis = PCA_ANM_Analysis()
    Analysis.set_default_based_on_argparse(selection, vmd, compare, ref, debug, pdb)
    
    pdbs = Analysis.getPDBs()
    ensemble = Analysis.createEnsemble(pdbs, logger)
    # set weights for the PCA analysis
    # ensemble = Analysis.setWeights(ensemble)
    ensemble = Analysis.setSelections(ensemble)
    
    pca, pca_modes = Analysis.calcPCA(ensemble, logger)
    anm = Analysis.calcANM(ensemble.getConformation(0), logger)
    
    if Analysis.compare:
        Analysis.compare_modes(anm,pca)
   
    return pca_modes
    
    
if __name__ == '__main__':
    # parse command line options and create help
    args = parse_Arguments()
    main(args.pdb, args.selection, args.vmd, args.compare, args.ref, args.debug)




import os
from dataclasses import dataclass
import glob
import shutil
import logging
import pele_platform.Utilities.Helpers.solventOBCParamsGenerator as obc

@dataclass
class ImplicitSolvent:
    
    solvent: str
    obc_tmp: str
    template_folder: str
    obc_file: str
    logger: logging.Logger

    def generate(self):
        self.logger.info("Setting implicit solvent: {}".format(self.solvent))
        self.set_implicit_solvent()
        self.logger.info("Implicit solvent set\n\n".format(self.solvent))

    def set_implicit_solvent(self):
        if self.solvent == "OBC": #OBC from ...
            shutil.copy(self.obc_tmp, self.obc_file)
            for template in glob.glob(os.path.join(self.template_folder, "*")):
                obc.main(template, self.obc_file)
        else:
            pass #SVG by default



import os
import shutil
import glob
import subprocess
import mdtraj
import numpy as np
import pandas as pd
from sklearn import mixture
from multiprocessing import Pool
import matplotlib.pyplot as plt
import AdaptivePELE.analysis.selectOnPlot as sp
import pele_platform.Utilities.Helpers.bestStructs as bs
from pele_platform.constants import constants as cs
import pele_platform.Analysis.pdf_report as pr
from pele_platform.Utilities.Helpers.helpers import backup_logger

EPOCH = "epoch"
STEPS = 3
TRAJECTORY = "trajectory"


def _extract_coords(info):
    p, v, resname, topology = info
    # Most time consuming step 0.1
    traj = mdtraj.load_frame(p, v, top=topology)
    atoms_info = traj.topology.to_dataframe()[0]
    condition = atoms_info['resName'] == resname
    atom_numbers_ligand = atoms_info[condition].index.tolist()
    coords = []
    for atom_num in atom_numbers_ligand:
        try:
            coords.extend(traj.xyz[0][atom_num].tolist())
        except IndexError:
            continue
    return np.array(coords).ravel()


class PostProcessor:

    def __init__(self, report_name, traj_name, simulation_path, cpus, topology=False, residue=False,
                 be_column=4, limit_column=6, te_column=3, logger=None):
        self.report_name = report_name
        self.traj_name = traj_name
        self.simulation_path = simulation_path
        self.data = self.retrive_data()
        self.topology = topology
        self.residue = residue
        self.be_column = be_column
        self.limit_column = limit_column
        self.te_column = te_column
        self.cpus = cpus
        self.logger = logger

    def retrive_data(self, separator=","):
        summary_csv_filename = os.path.join(self.simulation_path, "summary.csv")
        if not os.path.exists(summary_csv_filename):
            try:
                sp.concat_reports_in_csv(adaptive_results_path=self.simulation_path,
                                         output_file_path=summary_csv_filename,
                                         report_prefix=self.report_name, trajectory_prefix=self.traj_name,
                                         separator_out=separator)
            except ValueError:
                raise ValueError("Not report found under {}. Did you point to the right folder?".format(os.getcwd()))
        dataframe = pd.read_csv(summary_csv_filename, sep=separator, engine='python', header=0)
        dataframe_filtered = self._remove_outliers(dataframe)
        
        return dataframe_filtered

    def _remove_outliers(self, dataframe):

        cols = list(dataframe.columns)
        n_points_to_remove = int(len(dataframe[cols[0]]) * 0.02) # remove top 2% of each energy
        dataframe_filtered = dataframe.sort_values(cols[3], ascending=False).iloc[n_points_to_remove:]
        dataframe_filtered = dataframe_filtered.sort_values(cols[4], ascending=False).iloc[n_points_to_remove:]

        return dataframe_filtered

    def plot_two_metrics(self, column_to_x, column_to_y, column_to_z=False, output_name=None, output_folder="."):
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)

        column_to_x = column_to_x if not str(column_to_x).isdigit() else self._get_column_name(self.data, column_to_x)
        column_to_y = column_to_y if not str(column_to_y).isdigit() else self._get_column_name(self.data, column_to_y)
        
        fig, ax = plt.subplots()
        if column_to_z:
            column_to_z = column_to_z if not str(column_to_z).isdigit() else self._get_column_name(self.data, column_to_z)
            output_name = output_name if output_name else "{}_{}_{}_plot.png".format(column_to_x, column_to_y, column_to_z)
            output_name = os.path.join(output_folder,output_name).replace(" ", "_")
            
            pts = ax.scatter(self.data[column_to_x], self.data[column_to_y], c=self.data[column_to_z], s=20)
            cbar = plt.colorbar(pts)
            cbar.ax.set_ylabel(column_to_z)
            ax.set_xlabel(column_to_x)
            ax.set_ylabel(column_to_y)
            plt.savefig(output_name)
            backup_logger(self.logger, "Plotted {} vs {} vs {}".format(column_to_x, column_to_y, column_to_z))
        else:
            output_name = output_name if output_name else "{}_{}_plot.png".format(column_to_x, column_to_y)
            output_name = os.path.join(output_folder,output_name).replace(" ", "_")
            ax.scatter(self.data[column_to_x], self.data[column_to_y], s=20)
            ax.set_xlabel(column_to_x)
            ax.set_ylabel(column_to_y)
            plt.savefig(output_name)
            backup_logger(self.logger, "Plotted {} vs {}".format(column_to_x, column_to_y))
        return output_name

    def top_poses(self, metric, n_structs, output="BestStructs"):
        metric = metric if not str(metric).isdigit() else self._get_column_name(self.data, metric)
        best_poses = self.data.nsmallest(n_structs, metric)
        self._extract_poses(best_poses, metric, output)

    def _extract_poses(self, poses, metric, output):
        # Process data to be outputted
        values = poses[metric].tolist()
        self.best_energies = values
        paths = poses[TRAJECTORY].tolist()
        epochs = poses[EPOCH].tolist()
        file_ids = [traj[:-4].split("_")[-1] for traj in paths]
        steps = self._get_column_name(self.data, STEPS)
        step_indexes = poses[steps].tolist()
        out_freq = 1
        files_out = [
            "epoch{}_trajectory_{}.{}_{}{}.pdb".format(epoch, report, int(step), metric.replace(" ", ""), value) \
            for epoch, step, report, value in zip(epochs, step_indexes, file_ids, values)]

        # Read trajectory and output snapshot
        for f_id, f_out, step, path in zip(file_ids, files_out, step_indexes, paths):
            if not self.topology:
                try:
                    bs.extract_snapshot_from_pdb(path, f_id, output, self.topology, step, out_freq, f_out, self.logger)
                except UnicodeDecodeError:
                    raise Exception(
                        "XTC output being treated as PDB. Please specify XTC with the next flag. traj: 'trajectory_name.xtc' in the input.yaml")
            else:
                bs.extract_snapshot_from_xtc(path, f_id, output, self.topology, step, out_freq, f_out, self.logger)

    def cluster_poses(self, n_structs, metric, output, nclusts=10):
        assert self.residue, "Set residue ligand name to clusterize"
        metric = metric if not str(metric).isdigit() else self._get_column_name(self.data, metric)
        best_poses = self.data.nsmallest(n_structs, metric)
        clusters = self._cluster(best_poses, metric, output, nclusts)
        return clusters

    def _cluster(self, poses, metric, output, nclusts=10):
        # Extract metric values
        values = poses[metric].tolist()
        epochs = poses[EPOCH].tolist()
        paths = poses[TRAJECTORY].tolist()
        file_ids = [traj[:-4].split("_")[-1] for traj in paths]
        steps = self._get_column_name(self.data, STEPS)
        snapshots = poses[steps].tolist()

        # Extract coords
        pool = Pool(processes=self.cpus)
        input_pool = [[p, v, self.residue, self.topology] for p, v in zip(paths, snapshots)]
        all_coords = pool.map(_extract_coords, input_pool)

        # Clusterize
        assert all_coords[0][
            0], "Ligand not found check the option --resname. i.e python interactive.py 5 6 7 --resname LIG"
        try:
            clf = mixture.GaussianMixture(n_components=nclusts, covariance_type='full')
            labels = clf.fit_predict(all_coords)
            indexes = labels
        except ValueError:
            indexes = [1]
        n_clusters = len(set(indexes))
        files_out = ["cluster{}_epoch{}_trajectory_{}.{}_{}{}.pdb".format(cluster, epoch, report, int(step),
                                                                          metric.replace(" ", ""), value) \
                     for epoch, step, report, value, cluster in zip(epochs, snapshots, file_ids, values, indexes)]
        all_metrics = []
        output_clusters = []

        for n_cluster in range(n_clusters - 1):
            metrics = {value: idx for idx, (value, cluster) in enumerate(zip(values, indexes)) if n_cluster == cluster}
            out_freq = 1
            cluster_metrics = list(metrics.keys())
            max_idx = metrics[np.min(cluster_metrics)]
            max_traj = paths[max_idx]
            max_snapshot = snapshots[max_idx]
            output_traj = files_out[max_idx]
            input_traj = file_ids[max_idx]
            if not self.topology:
                bs.extract_snapshot_from_pdb(max_traj, input_traj, output, self.topology, max_snapshot, out_freq,
                                             output_traj, logger=self.logger)
            else:
                bs.extract_snapshot_from_xtc(max_traj, input_traj, output, self.topology, max_snapshot, out_freq,
                                             output_traj, logger=self.logger)
            all_metrics.append(cluster_metrics)
            output_clusters.append(os.path.join(output, output_traj))
        fig, ax = plt.subplots()
        try:
            ax.boxplot(all_metrics)
        except IndexError:
            logger.info("Samples to disperse to produce a cluster")
            return
        ax.set_ylabel(metric)
        ax.set_xlabel("Cluster number")
        plt.savefig(os.path.join(output, "clusters_{}_boxplot.png".format(metric)))
        return output_clusters

    def _get_column_name(self, df, column_digit):
        return list(df)[int(column_digit) - 1]


def analyse_simulation(report_name, traj_name, simulation_path, residue, output_folder=".", cpus=5, clustering=True,
                       mae=False, nclusts=10, overwrite=False, topology=False, be_column=4, limit_column=6, te_column=3,
                       logger=None):
    results_folder = os.path.join(output_folder, "results")
    if os.path.exists(results_folder):
        if not overwrite:
            raise ValueError(
                "Analysis folder {} already exists. Use the option overwrite_analysis: true".format(results_folder))
        else:
            shutil.rmtree(os.path.join(output_folder, "results"))
    analysis = PostProcessor(report_name, traj_name, simulation_path, cpus, residue=residue, topology=topology,
                             be_column=be_column, limit_column=limit_column, te_column=te_column, logger=logger)
    
    metrics = len(list(analysis.data)) - 1  # Discard epoch as metric
    be = analysis.be_column
    total_energy = analysis.te_column
    current_metric = analysis.limit_column
    plots_folder = os.path.join(output_folder, "results/Plots")
    top_poses_folder = os.path.join(output_folder, "results/BestStructs")
    clusters_folder = os.path.join(output_folder, "results/clusters")

    if not os.path.exists(plots_folder):
        os.makedirs(plots_folder)
    if not os.path.exists(top_poses_folder):
        os.makedirs(top_poses_folder)
    if not os.path.exists(clusters_folder):
        os.makedirs(clusters_folder)

    # Plot metrics
    while current_metric <= metrics - 1:
        try:
            analysis.plot_two_metrics(total_energy, be, current_metric, output_folder=plots_folder)
            analysis.plot_two_metrics(current_metric, be, output_folder=plots_folder)

        except ValueError:
            break
        current_metric += 1

    # Retrieve 100 best structures
    logger.info("Retrieve 100 Best Poses")
    analysis.top_poses(be, 100, top_poses_folder)

    # Clustering of best 2000 best structures
    logger.info(f"Retrieve {nclusts} best cluster poses")
    if clustering:
        clusters = analysis.cluster_poses(250, be, clusters_folder, nclusts=nclusts)
    if mae:
        sch_python = os.path.join(cs.SCHRODINGER, "utilities/python")
        if not os.path.exists(sch_python):
            sch_python = os.path.join(cs.SCHRODINGER, "run")
        top_poses = glob.glob(os.path.join(top_poses_folder, "*"))
        python_file = os.path.join(cs.DIR, "Analysis/to_mae.py")
        for poses in top_poses + clusters:
            command = "{} {} {} --schr {} {}".format(sch_python, python_file, poses, cs.SCHRODINGER, "--remove")
            logger.info(command)
            try:
                subprocess.check_call(command.split())
            except ValueError:
                raise ValueError(
                    "Binding energy is not in the default report column (4). Please specify the column by means of the next flag. be_column: N")
    plots = glob.glob(os.path.join(plots_folder, "*.png"))
    poses = glob.glob(os.path.join(top_poses_folder, "*"))
    clusters = glob.glob(os.path.join(clusters_folder, "*.png"))
    report = pr.create_report(plots, clusters, poses, analysis.best_energies,
                              output=os.path.join(output_folder, "summary_results.pdf"))
    return report



from dataclasses import dataclass, field
from typing import List
from Bio.PDB import PDBParser, PDBIO, Selection, NeighborSearch, Vector
import glob
import numpy as np
import os
import pele_platform.constants.constants as cs
import pele_platform.Errors.custom_errors as ce
import pele_platform.Utilities.Helpers.helpers as hp
import pele_platform.constants.pele_params as pp

TEMPLATE = '{{"watersToPerturb": {{"links": {{"ids": [{index}] }}}}, "Box": {{"radius": {radius}, "fixedCenter": [{com}], "type": "sphericalBox"}}}}'

@dataclass
class WaterIncluder():

    input_pdbs: list
    n_waters: int
    user_waters: List=field(default_factory=lambda: [])
    ligand_perturbation_params: str=""
    ligand_residue: str=""
    water_center: bool=False
    water_radius: bool=False
    water_to_exclude: List=field(default_factory=lambda: [])
    sim_path: str="."
    allow_empty_selectors: bool=False
    water_temp: int=5000
    water_trials: int=10000
    water_overlap: float=0.78
    water_constr: float=0
    water_freq: int=1
    all_waters: str=""
    perturbed_waters=[]
    test: bool=False

    
    def run(self):
        self.set_empty_selectors()
        self.set_user_waters()
        self.water_checker()
        self.add_water()
        self.set_water_control_file()
    
    def set_empty_selectors(self):
        #If True include in control file else null
        self.allow_empty_selectors = '"allowEmptyWaterSelectors": true,' if self.allow_empty_selectors else ""

    def set_user_waters(self):
        #If all_waters extract them from the any pdb
        if self.user_waters == "all_waters":
            self.user_waters = hp.retrieve_all_waters(self.input_pdbs[0])
        
        
    def set_parameters(self):
        if self.ligand_perturbation_params:
            self.ligand_perturbation_params = self.ligand_perturbation_params.rstrip("]\n") + pp.WATER_PARAMS
        else:
            self.ligand_perturbation_params = ', "parametersChanges" : [ ' + pp.WATER_PARAMS.replace(",{", "{")

    def set_box_center(self, inp=False):
        if self.ligand_residue:
            com = ligand_com(inp, self.ligand_residue)
            com_format = ['{:.10f}'.format(elem) for elem in com[0]]
            self.water_center = ", ".join(com_format)
        else:
            com_format = ['{:.10f}'.format(coord) for coord in self.water_center]
            self.water_center = ", ".join(com_format)


    def set_box_radius(self):
        if self.water_radius:
            self.water_radius =  self.water_radius
        else:
            self.water_radius = 6

    def set_water_input(self, inp):
        try:
            self.set_box_center(inp)
        except TypeError:
            raise ce.NotCenterOfWaterBox('Center of water box not found. Specify with the next flag\n\n"water_center":\n - 18\n - 17\n - 18')
        self.set_box_radius()
        waters = sorted(hp.retrieve_all_waters(inp, exclude=self.water_to_exclude))
        waters = ", ".join(['"{}"'.format(water) for water in waters])
        self.all_waters += waters + ", "
        return TEMPLATE.format(index=waters, radius=self.water_radius, com=self.water_center)
             

    def set_water_control_file(self):
        if self.n_waters != 0 or self.user_waters:
            self.set_parameters()
            water_string = [self.set_water_input(inp).strip("'") for inp in self.input_pdbs]
            self.all_waters = self.all_waters.rstrip(", ")
            self.water_line = cs.WATER.format(self.all_waters, self.allow_empty_selectors, self.water_temp, self.water_trials,
                                        self.water_overlap, self.water_constr, water_string).replace("'", "")
            self.water_energy = None  # TEMPORARY FIX
    
        else:
            self.water_energy = None
            self.water_radius = None
            self.water_center = None
            self.water_line = ""
    
    
    def water_checker(self):
        max_water = 4
        min_water = 1
        if self.n_waters:
            if self.n_waters > max_water:
                raise ValueError("Maximum {} water molecules are allowed.".format(max_water))
            elif self.n_waters < min_water:
                raise ValueError(
                    "Number of water molecules (n_waters) has to be between {} and {}".format(min_water, max_water))
    
    
    def add_water(self):
        if self.test:
            np.random.seed(42)
    
    
        output = []
        n_inputs = len(self.input_pdbs)
        water_coords = []
        resnums = []
        atomnums = []
        chains = []
        resnames = []
    
        # get maximum residue and atom numbers keep original waters
        with open(self.input_pdbs[0], "r") as file:
            pdb_lines = [ line for line in file.readlines() if "END" not in line]
            for line in pdb_lines:
                if line.startswith("ATOM") or line.startswith("HETATM") or line.startswith("TER"):
                    try:
                        resnum = line[22:27].strip()
                        atomnum = line[7:11].strip()
                        chain = line[21]
                        resname = line[17:20]
                        resnums.append(resnum)
                        atomnums.append(atomnum)
                        chains.append(chain)
                        resnames.append(resname)
                        if resname == "HOH":
                            water = f"{chain}:{resnum}"
                            if water not in self.user_waters and water not in self.water_to_exclude:
                                self.water_to_exclude.append(water)
                    #Line too short - Remarks pdb
                    except IndexError:
                        pass

        if self.n_waters < 1:
            return
        else:
            lig_length = resnames.count(self.ligand_residue)
            resnums = [int(num) for num in resnums if num]
            max_resnum = max(resnums)
            water_resnums = []
            water_chain = chains[0]  # water chain = 1st protein chain
            atomnum = max([int(num) for num in atomnums if num]) + 1 + lig_length
    
            water = cs.water * self.n_waters * n_inputs
            for inp in range(n_inputs):
                for n in range(self.n_waters):
                    O_coords = Vector([np.random.randint(0, 100) for i in range(3)])
                    H1_coords = O_coords + Vector(0.757, 0.586, 0.0)
                    H2_coords = O_coords + Vector(-0.757, 0.586, 0.0)
                    water_coords = water_coords + [list(O_coords)] + [list(H1_coords)] + [list(H2_coords)]
    
                    max_resnum += 1  # each water must have a different residue number
                    water_resnums = water_resnums + [max_resnum] * 3
                max_resnum += 1
    
            water_atomnums = [atomnum + j for j in range(self.n_waters * 3 * n_inputs)]
    
            # PDB lines - water
            water_output = []
            for atom, num, resnum, coord in zip(water, water_atomnums, water_resnums, water_coords):
                coord = ["{:7.4f}".format(c) for c in coord]
                coord = " ".join(coord)
                water_output.append(atom.format(num, water_chain, resnum, coord))
    
            sliced_water_output = []
            for i in range(0, len(water_output), self.n_waters * 3):
                sliced_water_output.append(water_output[i:i + self.n_waters * 3])
    
            # loop over minimisation inputs
            for inp, w in zip(self.input_pdbs, sliced_water_output):
                new_protein_file = inp
                # add water to PDB
                with open(inp, "w+") as file:
                    for line in pdb_lines:
                        file.write(line)
                    file.write("\n")
                    for line in w:
                        file.write(line)
                    file.write("END")
    
                # load again with Biopython
                parser = PDBParser()
                structure = parser.get_structure("complex", new_protein_file)
                water_list = []
                protein_list = Selection.unfold_entities(structure, "A")
                temp_protein_file = os.path.join(os.path.dirname(inp),
                                                 os.path.basename(inp).replace(".pdb", "_temp.pdb"))
    
                for res in structure.get_residues():
                    resnum = res._id[1]
                    if res.resname == 'HOH':
                        if resnum not in resnums:
                            water_list = water_list + Selection.unfold_entities(res, "A")
    
                # check for water contacts
                contacts5 = []
                for w in water_list:
                    contacts5 = contacts5 + NeighborSearch(protein_list).search(w.coord, 5.0, "A")
                contacts5 = [c for c in contacts5 if c not in water_list]  # exclude "self" contacts
    
                # translate water, if needed
                while contacts5:
                    contacts5 = []
                    for w_ in water_list:
                        x, y, z = w_.coord
                        w_.set_coord([x - 5, y, z])
                        contacts5 = contacts5 + NeighborSearch(protein_list).search(w_.coord, 5.0, "A")
                        contacts5 = [c for c in contacts5 if c not in water_list]
    
                # save final output
                io = PDBIO()
                io.set_structure(structure)
                io.save(temp_protein_file)
                output.append(new_protein_file)
    
                new_water_lines = []
                with open(temp_protein_file, "r") as temp:
                    temp_lines = temp.readlines()
                    for line in temp_lines:
                        if line[17:20].strip() == "HOH" and int(line[22:27].strip()) not in resnums:
                            line = line.replace(line[7:11], str(int(line[7:11]) + lig_length))
                            if line[12:15] == "2HW":
                                line = line.strip("\n") + "\nTER\n"
                            new_water_lines.append(line)
    
                del new_water_lines[-1] #Last biopython line is a not need it TER
                
                with open(new_protein_file, "w+") as file:
                    for line in pdb_lines:
                        file.write(line)
                    if not line.startswith("TER"):
                        file.write("TER\n")
                    for line in new_water_lines:
                        file.write(line)
                    file.write("END")
    
                os.remove(temp_protein_file)


def ligand_com(refinement_input, ligand_chain):
    parser = PDBParser()
    output = []
    refinement_input = glob.glob(refinement_input)

    for inp in refinement_input:
        structure = parser.get_structure("inp", inp)
        mass = 0.0
        com = np.zeros(3)
        for res in structure.get_residues():
            if res.resname == ligand_chain:
                for atom in res.get_atoms():
                    com = com + np.array(list(atom.get_vector())) * atom.mass
                    mass += atom.mass
                com = com / mass

        output.append(com.tolist())

    return output



import os
import pandas as pd
from sklearn.mixture import GaussianMixture
from pele_platform.Utilities.Helpers import bestStructs as bs
from pele_platform.Analysis.plots import _extract_coords
from multiprocessing import Pool



def cluster_best_structures(be_column: int, residue="LIG", topology=None, cpus=20, n_components=10, n_structs=1000, directory=".", logger=None):

    files_out, _, _, _, output_energy = bs.main(be_column, n_structs=n_structs, path=".", topology=topology, logger=logger)
    n_files = len(files_out)
    logger.info("Extracting data from {} files.".format(n_files))

    snapshot = 0
    pool = Pool(cpus)
    input_pool = [[f, snapshot, residue, topology] for f in files_out]  
    all_coords = pool.map(_extract_coords, input_pool)

    # cluster
    logger.info("Creating clusters.")
    cluster = GaussianMixture(n_components, covariance_type='full', random_state=42)
    labels = cluster.fit_predict(all_coords)

    # get lowest energy representative
    clustered_lig = pd.DataFrame(list(zip(files_out, labels, output_energy)), columns=["file_name", "cluster_ID", "binding_energy"])
    clustered_lig = clustered_lig.sort_values("binding_energy", ascending=True).groupby("cluster_ID").first()
    clustered_lig.to_csv("clustering_output.csv")

    output_files = clustered_lig["file_name"].values.tolist()
    directory = os.path.join(directory, "refinement_input")

    # copy selected files to create input for refinement
    if not os.path.isdir(directory):
        os.makedirs(directory, exist_ok=True)
    for file in output_files:
        os.system("cp {} {}/.".format(file, directory))

    return clustered_lig['file_name']



import os
import pele_platform.constants.constants as cs
import pele_platform.constants.pele_params as pcs

"""
Description of the file: This file specifies what files to use
for each job.

"""
    
def retrieve_software_settings(args, pele_dir):

        SOFTWARE_CONSTANTS = {
                 "adap_ex_output" : None,
                 "ad_ex_temp" : os.path.join(pele_dir, "adaptive.conf"),
                 "pele_exit_temp" : os.path.join(pele_dir, "pele.conf"),
                 "folders" : [ "",
                    "DataLocal/Templates/OPLS2005/HeteroAtoms/",
                    "DataLocal/Templates/AMBER99sb/HeteroAtoms/",
                    "DataLocal/Templates/AMBER99sbBSC0/HeteroAtoms/",
                    "DataLocal/LigandRotamerLibs",
                    "DataLocal/OBC"
                    ],
                  "file_names" : ["adaptive.conf", "pele.conf" ],
                  "files" : [os.path.join(cs.DIR, "Templates/template_adaptive.conf"), 
                             os.path.join(cs.DIR, "Templates/pele_template.conf") ],
                 "simulation_params" : {
                             "out_in": {"spawning_type": "inverselyProportional", "density": "continuous",
                                      "simulation_type": "pele", "iterations": 100, "pele_steps": 8, 
                                      "cluster_values": "[2, 5, 7]", "cluster_conditions": "[1, 0.6, 0.0]",
                                      "steric_trials": 250, "overlap_factor": 0.65, "params": pcs.OUT_IN,
                                      "box_radius": 30},
                             "gpcr_orth": {"spawning_type": "epsilon", "density": "null",
                                      "simulation_type": "pele", "iterations": 50, "pele_steps": 8, 
                                      "cluster_values": "[1.75, 2.5, 4]", "cluster_conditions": "[0.7, 0.4, 0.0]",
                                      "steric_trials": 100, "overlap_factor": 0.65, "epsilon": 0.25,
                                      "bias_column": 6, "params": pcs.GPCR_ORTH, "box_radius": 15, "ca_constr": 5},
                             "global": {"spawning_type": "inverselyProportional", "bias_column": 5, "epsilon":0.25, "density": "continuous",
                                      "simulation_type": "pele", "iterations": 100, "pele_steps": 4, 
                                      "cluster_values": "[2.5, 5, 7]", "cluster_conditions": "[1, 0.6, 0.0]",
                                      "steric_trials": 200, "overlap_factor": 0.65, "params": pcs.GLOBAL,
                                      "box_radius":None},
                             "induced_fit_exhaustive" : {"spawning_type": "independent", "bias_column": 5, "epsilon":0.25, "density": "null",
                                      "simulation_type": "pele", "iterations": 1, "pele_steps": 1000, 
                                      "cluster_values": "[1.75, 2.5, 4, 6]", "cluster_conditions": "[1, 0.6, 0.4, 0.0]",
                                      "steric_trials": 500, "overlap_factor": 0.65, "params": pcs.INDUCED_FIT,
                                      "box_radius": 6},
                             "induced_fit_fast" : {"spawning_type": "inverselyProportional", "bias_column": 5, "epsilon":0.25, "density": "null",
                                      "simulation_type": "pele", "iterations": 30, "pele_steps": 12, 
                                      "cluster_values": "[1.75, 2.5, 4, 6]", "cluster_conditions": "[1, 0.6, 0.4, 0.0]",
                                      "steric_trials": 500, "overlap_factor": 0.65, "params": pcs.INDUCED_FIT,
                                      "box_radius": 6},
                             "in_out" : {"spawning_type": "epsilon", "bias_column": 6, "epsilon":0.75, "density": "exitContinuous",
                                      "simulation_type": "pele", "iterations": 1000, "pele_steps": 2, 
                                      "cluster_values": "[1, 2.5]", "cluster_conditions": "[1.1]",
                                      "steric_trials": 500, "overlap_factor": 0.65, "params": pcs.IN_OUT,
                                      "box_radius": 10},
                             "in_out_soft" : {"spawning_type": "independentMetric", "bias_column": 6, "epsilon":0.75, "density": "exitContinuous",
                                      "simulation_type": "pele", "iterations": 1000, "pele_steps": 2, 
                                      "cluster_values": "[1, 2.5]", "cluster_conditions": "[1.1]",
                                      "steric_trials": 500, "overlap_factor": 0.65, "params": pcs.IN_OUT,
                                      "box_radius": 10, "spawning_condition": "max"},
                             "rescoring" : {"spawning_type": "independent", "bias_column": 5, "epsilon":0.25, "density": "null",
                                      "simulation_type": "pele", "iterations": 20, "pele_steps": 12,
                                      "cluster_values": "[1.75, 2.5, 4, 6]", "cluster_conditions": "[1, 0.6, 0.4, 0.0]",
                                      "steric_trials": 500, "overlap_factor": 0.65, "params": pcs.RESCORING,
                                      "box_radius": 6, "anm_freq": 6, "sidechain_freq": 3, "min_freq": 1, "temperature": 1000,
                                      "anm_displacement": 0.5, "anm_modes_change": 3},
                             "anm": {"spawning_type": "independent", "bias_column": 5, "epsilon":0.15, "density": "null",
                                      "simulation_type": "pele", "iterations": 50, "pele_steps": 8, 
                                      "cluster_values": "[1.5, 2, 5]", "cluster_conditions": "[0.6, 0.4, 0.0]",
                                      "steric_trials": 250, "overlap_factor": 0.65, "params": pcs.OUT_IN,
                                      "box_radius": 30}
                    }

        }
        
        software_setings = SOFTWARE_CONSTANTS
        if args.full:
            type_simulation = "global"
        elif args.in_out:
            type_simulation = "in_out"
        elif args.in_out_soft:
            type_simulation = "in_out_soft"
        elif args.induced_fit_exhaustive:
            type_simulation = "induced_fit_exhaustive"
        elif args.induced_fit_fast:
            type_simulation = "induced_fit_fast"
        elif args.rescoring:
            type_simulation = "rescoring"
        elif args.out_in:
            type_simulation = "out_in"
        elif args.gpcr_orth:
            type_simulation = "gpcr_orth"
        else:
            #Standard file (user will change the parameters)
            type_simulation = "induced_fit_fast"

        software_setings["files"] = software_setings["files"]
        software_setings["simulation_params"] = software_setings["simulation_params"].get(type_simulation)
        return software_setings



from dataclasses import dataclass
import os
import pele_platform.Errors.custom_errors as ce


@dataclass
class Executable():


    executable: str

    def is_in_path(self) -> bool:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, self.executable)
            if self._is_exe(exe_file):
                return True
        raise ce.ExecutableNotInPath(f"Executable {self.executable} not in PATH. Please do: export PATH=/path/to/folder/with/executable:$PATH")

    def _is_exe(self, fpath) -> str:
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)



import matplotlib
matplotlib.use("Agg") #Avoid backend issues
import pele_platform.Checker.python_version as pv
pv.check_python_version() #Avoid python2
from argparse import ArgumentParser
import pele_platform.Utilities.Helpers.yaml_parser as yp
import pele_platform.Utilities.Helpers.launcher as lc
import pele_platform.Utilities.Parameters.pele_env as pv
import pele_platform.Checker.valid_flags as vf
import pele_platform.Errors.custom_errors as ce



def parseargs(args=[]):
    '''
    Command line parser
    '''
    parser = ArgumentParser(description='Run PELE Platform')
    parser.add_argument('input_file', type=str, help='Yaml input file')
    args = parser.parse_args(args) if args else parser.parse_args()
    return args.input_file
    
def run_platform(input_yaml):
    '''
    High level function to run a PELE job. It will:
    1) Parse the input.yaml
    2) Launch job
    3) Return job parametrs
    '''
    yaml_obj = yp.YamlParser(input_yaml, vf.VALID_FLAGS_PLATFORM)
    try:
        yaml_obj.read()
    except AttributeError:
        raise ce.WrongYamlFile("Input file: {} does not look like a correct yml file".format(input_yaml))
    job_params = lc.Launcher(yaml_obj).launch()
    return job_params

if __name__ == "__main__":
    input_file = parseargs()
    job = run_platform(input_file)



import os
from pele_platform.Allosteric.cluster import cluster_best_structures
from pele_platform.Utilities.Helpers.helpers import cd, is_repited, is_last
import pele_platform.Utilities.Parameters.pele_env as pv
import pele_platform.Adaptive.simulation as si


def run_allosteric(parsed_yaml: dict) -> (pv.EnviroBuilder, pv.EnviroBuilder):
    """
    Run allosteric simulation by:
    1) Run global exploration to identify most important pockets
    2) Run induced fit simulation to find deep pockets
    """

    # Let user choose working folder
    original_dir = os.path.abspath(os.getcwd())
    working_folder = os.path.abspath("{}_Pele".format(parsed_yaml.residue))
    if not parsed_yaml.folder:
        working_folder = is_repited(working_folder) if not parsed_yaml.adaptive_restart else is_last(working_folder)
    else:
        working_folder = os.path.abspath(parsed_yaml.folder)

    # Set main folder
    parsed_yaml.folder = os.path.join(working_folder, "1_global_exploration")

    # start initial simulation
    parsed_yaml.full = True
    simulation = si.run_adaptive(parsed_yaml)
    simulation_path = os.path.join(simulation.pele_dir, simulation.output)
    
    # get best structures and cluster them
    if not parsed_yaml.debug:
        with cd(simulation_path):
            # NEED ALGORITHM TO CHOOSE OPTIMUM NUMBERS OF CLUSTERS!!!!
            cluster_best_structures("5", n_components=simulation.n_components,
                residue=simulation.residue, topology=simulation.topology,
                directory=working_folder, logger=simulation.logger)
    
    # adjust original input.yaml
    if not parsed_yaml.skip_refinement:
        parsed_yaml.system = os.path.join(working_folder, "refinement_input/*.pdb")
        parsed_yaml.folder = os.path.join(working_folder, "2_refinement_simulation")
        parsed_yaml.full = None
        parsed_yaml.poses = None
        parsed_yaml.induced_fit_exhaustive = True

        if not parsed_yaml.test:
            parsed_yaml.iterations = 20
            parsed_yaml.pele_steps = 10
        parsed_yaml.box_center = simulation.box_center
        parsed_yaml.box_radius = simulation.box_radius

        # refine selected best structures
        with cd(original_dir):
            if not parsed_yaml.debug:
                induced_fit = si.run_adaptive(parsed_yaml)
            else:
                induced_fit = None
    else:
        induced_fit = None

    return simulation, induced_fit




import os
import argparse
import schrodinger.structure as st

####Here because otherwise need to charge helpers with PPP
####and schrodinger does not have that module
class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)

def pdb_to_mae(fname, schr_path, mae_output_file=None, remove=False):
    directory = os.path.dirname(fname)
    fname = os.path.basename(fname)
    with cd(directory):
        file_info = fname.split("_")
        binding_energy = float(file_info[-1].replace("BindingEnergy", "").replace(".pdb", ""))
        trajectory = int(file_info[-2].split(".")[0])
        snapshot = int(file_info[-2].split(".")[1])
        epoch = int(file_info[-4].split("epoch")[-1])
        properties = {"BindingEnergy": binding_energy,
           "trajectory": trajectory, 
           "snapshot": snapshot,
           "epoch": epoch
           }
        title = fname.split('_')
        traj = f"{epoch}_{trajectory}_{snapshot}"
        pele_energy = binding_energy
        cmd = '$SCHRODINGER/utilities/prepwizard -rehtreat -noepik -disulfides -noimpref -nometaltreat -noprotassign -nopropka -WAIT %s %s.mae' %(fname, traj)
        print(cmd)
        os.system(cmd)
        struct = next(st.StructureReader(f"{traj}.mae"))
        struct.property['r_user_PELE_energy'] = properties["BindingEnergy"]
        struct.property['r_user_PELE_epoch'] = properties["epoch"]
        struct.property['r_user_PELE_traj'] = properties["trajectory"]
        struct.property['r_user_PELE_snapshot'] = properties["snapshot"]
        struct.title = '%s_BindEn_%.2f' %(traj, pele_energy) 
        struct.write(f"{traj}.mae")


def add_args(parser):
    parser.add_argument('inputfile', type=str, help="Pdb input file")
    parser.add_argument('--schr', type=str, help="Schrodinger root path")
    parser.add_argument('--remove', action="store_true", help="Remove input file at exit")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Write MAE and SDF files with certain properties')
    add_args(parser)
    args = parser.parse_args()
    pdb_to_mae(args.inputfile, args.schr, remove=args.remove)




class MSMParams(object):


    def generate_msm_params(self, args):
        self.lagtime = args.lagtime
        self.msm_clust = args.msm_clust
        self.sasa_max = None
        self.sasa_min = None
        self.clusters = args.clust = args.clust if not args.test else 2
